#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
sendline, expect_prompt, expect_str = sp.sendline, sp.expect_prompt, sp.expect_str

# Test fish_postexec and $status_generation for interactive shells.
expect_prompt()

sendline(
    "function test_fish_postexec --on-event fish_postexec; printf 'pipestatus:%s, generation:%d, command:%s\\n' (string join '|' $pipestatus) $status_generation $argv; end"
)
expect_prompt()

generation = 1

# fish_postexec is called when foreground job ends.
sendline("true")
generation += 1
expect_str("pipestatus:0, generation:%d, command:true" % generation)
expect_prompt()

# Command has multiple jobs.
sendline("true;false")
generation += 1
expect_str("pipestatus:1, generation:%d, command:true;false" % generation)
expect_prompt()

# Command is a pipeline.
sendline("true|false")
generation += 1
expect_str("pipestatus:0|1, generation:%d, command:true|false" % generation)
expect_prompt()

# Does not increment $status_generation for background jobs.
# status, pipestatus remain unchanged
sendline("sleep 1000 &")
expect_str("pipestatus:0|1, generation:%d, command:sleep 1000 &" % generation)
expect_prompt()

# multiple backgrounded jobs
sendline("sleep 1000 &; sleep 2000 &")
expect_str(
    "pipestatus:0|1, generation:%d, command:sleep 1000 &; sleep 2000 &" % generation
)
expect_prompt()

# valid variable assignment
sendline("set foo bar")
expect_str("pipestatus:0|1, generation:%d, command:set foo bar" % generation)
expect_prompt()

# valid variable assignment with background job
sendline("set foo bar; sleep 1000 &")
expect_str(
    "pipestatus:0|1, generation:%d, command:set foo bar; sleep 1000 &" % generation
)
expect_prompt()

# Increments $status_generation if any job was foreground.
sendline("false|true; sleep 1000 &")
generation += 1
expect_str(
    "pipestatus:1|0, generation:%d, command:false|true; sleep 1000 &" % generation
)
expect_prompt()

sendline("sleep 1000 &; true|false|true")
generation += 1
expect_str(
    "pipestatus:0|1|0, generation:%d, command:sleep 1000 &; true|false|true"
    % generation
)
expect_prompt()

# Increments $status_generation for empty if/while blocks.
sendline("if true;end")
generation += 1
expect_str("pipestatus:0, generation:%d, command:if true;end" % generation)
expect_prompt()

sendline("while false;end")
generation += 1
expect_str("pipestatus:0, generation:%d, command:while false;end" % generation)
expect_prompt()

# or non-matching if.
sendline("if false;false;end")
generation += 1
expect_str("pipestatus:0, generation:%d, command:if false;false;end" % generation)
expect_prompt()

# or a function definition.
# This is an implementation detail, but the test case should prevent regressions.
sendline("function fail; false; end")
generation += 1
expect_str(
    "pipestatus:0, generation:%d, command:function fail; false; end" % generation
)
expect_prompt()

# or an invalid variable assignment
sendline("set '!@#$' value")
generation += 1
expect_str("pipestatus:2, generation:%d, command:set '!@#$' value" % generation)
expect_prompt()

# or a variable query
sendline("set -q fish_pid")
generation += 1
expect_str("pipestatus:0, generation:%d, command:set -q fish_pid" % generation)
expect_prompt()

# This is just to set a memorable pipestatus.
sendline("true|false|true")
generation += 1
expect_str("pipestatus:0|1|0")
expect_prompt()

# Does not increment $status_generation for empty begin/end block.
sendline("begin;end")
expect_str("pipestatus:0|1|0, generation:%d, command:begin;end" % generation)
expect_prompt()

# Or begin/end block with only backgrounded jobs.
sendline("begin; sleep 200 &; sleep 400 &; end")
expect_str(
    "pipestatus:0|1|0, generation:%d, command:begin; sleep 200 &; sleep 400 &; end"
    % generation
)
expect_prompt()

# Or a combination of begin/end block and backgrounded job.
sendline("begin; sleep 200 &; end; sleep 400 &")
expect_str(
    "pipestatus:0|1|0, generation:%d, command:begin; sleep 200 &; end; sleep 400 &"
    % generation
)
expect_prompt()

# Or a combination with variable assignments
sendline("begin; set foo bar; sleep 1000 &; end; set bar baz; sleep 2000 &")
expect_str(
    "pipestatus:0|1|0, generation:%d, command:begin; set foo bar; sleep 1000 &; end; set bar baz; sleep 2000 &"
    % generation
)
expect_prompt()



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
import os

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)

# We're going to use three history files, including the default, to verify
# that the fish_history variable works as expected.
# (using as a relative path to avoid interpolating $XDG_DATA_HOME)
default_histfile = "xdg_data_home/fish/fish_history"
my_histfile = "xdg_data_home/fish/my_history"
env_histfile = "xdg_data_home/fish/env_history"


def grephistfile(line, file):
    sendline("grep '^" + line + "' " + file)


# Verify that if we spawn fish with no fish_history env var it uses the
# default file.
expect_prompt()

# Verify that a command is recorded in the default history file.
cmd1 = "echo $fish_pid default histfile"
hist_line = "- cmd: " + cmd1
sendline(cmd1)
expect_prompt()

# TODO: Figure out why this `history --save` is only needed in one of the
# three Travis CI build environments and neither of my OS X or Ubuntu servers.
sendline("history --save")
expect_prompt()

grephistfile(hist_line, default_histfile)
expect_str(hist_line)
expect_prompt()

# Switch to a new history file and verify it is written to and the default
# history file is not written to.
cmd2 = "echo $fish_pid my histfile"
hist_line = "- cmd: " + cmd2
sendline("set fish_history my")
expect_prompt()
sendline(cmd2)
expect_prompt()

grephistfile(hist_line, my_histfile)
expect_str(hist_line)
expect_prompt()

# We expect this grep to fail to find the pattern and thus the expect_prompt
# block is inverted.
grephistfile(hist_line, default_histfile)
expect_prompt()

# Switch back to the default history file.
cmd3 = "echo $fish_pid default histfile again"
hist_line = "- cmd: " + cmd3
sendline("set fish_history default")
expect_prompt()
sendline(cmd3)
expect_prompt()

# TODO: Figure out why this `history --save` is only needed in one of the
# three Travis CI build environments and neither of my OS X or Ubuntu servers.
sendline("history --save")
expect_prompt()

grephistfile(hist_line, default_histfile)
expect_str(hist_line)
expect_prompt()

# We expect this grep to fail to find the pattern and thus the expect_prompt
# block is inverted.
sendline("grep '^" + hist_line + "' " + my_histfile)
expect_prompt()
grephistfile(hist_line, my_histfile)
expect_prompt()

# Verify that if we spawn fish with a HISTFILE env var it uses that file.
# Start by shutting down the previous shell.
sendline("jobs")
expect_prompt("jobs: There are no jobs")
sendline("exit")
sendline("exit")
sp.spawn.wait()

# Set the fish_history env var.
os.environ["fish_history"] = "env"

# Spawn a new shell.
sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

# Verify that the new fish shell is using the fish_history value for history.
cmd4 = "echo $fish_pid env histfile"
hist_line = "- cmd: " + cmd4
sendline(cmd4)
expect_prompt()

grephistfile(hist_line, env_histfile)
expect_str(hist_line)
expect_prompt()

# We expect this grep to fail to find the pattern and thus the expect_prompt
# block is inverted.
grephistfile(hist_line, default_histfile)
expect_prompt()



# Pygments lexer for a fish command synopsis.
#
# Example usage:
# echo 'string match [OPTIONS] [STRING]' | pygmentize -f terminal256 -l doc_src/fish_synopsis.py:FishSynopsisLexer -x

from docutils import nodes
from pygments.lexer import Lexer
from pygments.token import (
    Generic,
    Name,
    Operator,
    Punctuation,
    Text,
)
import re
from sphinx.directives.code import CodeBlock


class FishSynopsisDirective(CodeBlock):
    """A custom directive that describes a command's grammar."""

    has_content = True
    required_arguments = 0

    def run(self):
        if self.env.app.builder.name != "man":
            self.arguments = ["fish-synopsis"]
            return CodeBlock.run(self)
        lexer = FishSynopsisLexer()
        result = nodes.line_block()
        for start, tok, text in lexer.get_tokens_unprocessed("\n".join(self.content)):
            if (  # Literal text.
                (tok in (Name.Function, Name.Constant) and not text.isupper())
                or text.startswith("-")  # Literal option, even if it's uppercase.
                or tok in (Operator, Punctuation)
                or text
                == " ]"  # Tiny hack: the closing bracket of the test(1) alias is a literal.
            ):
                node = nodes.strong(text=text)
            elif (
                tok in (Name.Constant, Name.Function) and text.isupper()
            ):  # Placeholder parameter.
                node = nodes.emphasis(text=text)
            else:  # Grammar metacharacter or whitespace.
                node = nodes.inline(text=text)
            result.append(node)
        return [result]


lexer_rules = [
    (re.compile(pattern), token)
    for pattern, token in (
        # Hack: treat the "[ expr ]" alias of builtin test as command token (not as grammar
        # metacharacter).  This works because we write it without spaces in the grammar (like
        # "[OPTIONS]").
        (r"\. |! |\[ | \]|\{ | \}", Name.Constant),
        # Statement separators.
        (r"\n", Text.Whitespace),
        (r";", Punctuation),
        (r" +", Text.Whitespace),
        # Operators have different highlighting than commands or parameters.
        (r"\b(and|not|or|time)\b", Operator),
        # Keywords that are not in command position.
        (r"\b(if|in)\b", Name.Function),
        # Grammar metacharacters.
        (r"[()[\]|]", Generic.Other),
        (r"\.\.\.", Generic.Other),
        # Parameters.
        (r"[\w-]+", Name.Constant),
        (r"[=%]", Name.Constant),
        (
            r"[<>]",
            Name.Constant,
        ),  # Redirection are highlighted like parameters by default.
    )
]


class FishSynopsisLexer(Lexer):
    name = "FishSynopsisLexer"
    aliases = ["fish-synopsis"]

    is_before_command_token = None

    def next_token(self, rule: str, offset: int, has_continuation_line: bool):
        for pattern, token_kind in lexer_rules:
            m = pattern.match(rule, pos=offset)
            if m is None:
                continue
            if token_kind is Name.Constant and self.is_before_command_token:
                token_kind = Name.Function

            if has_continuation_line:
                # Traditional case: rules with continuation lines only have a single command.
                self.is_before_command_token = False
            else:
                if m.group() in ("\n", ";") or token_kind is Operator:
                    self.is_before_command_token = True
                elif token_kind in (Name.Constant, Name.Function):
                    self.is_before_command_token = False

            return m, token_kind, m.end()
        return None, None, offset

    def get_tokens_unprocessed(self, input_text):
        """Return a list of (start, tok, value) tuples.

        start is the index into the string
        tok is the token type (as above)
        value is the string contents of the token
        """
        """
        A synopsis consists of multiple rules.  Each rule can have continuation lines, which
        are expected to be indented:

            cmd foo [--quux]
                    [ARGUMENT] ...
            cmd bar

        We'll split the input into rules. This is easy for a traditional synopsis because each
        non-indented line starts a new rule.  However, we also want to support code blocks:

            switch VALUE
               [case [GLOB ...]
                   [COMMAND ...]]
            end

        which makes this format ambiguous. Hack around this by always adding "end" to the
        current rule, which is enough in practice.
        """
        rules = []
        rule = []
        for line in list(input_text.splitlines()) + [""]:
            if rule and not line.startswith(" "):
                rules.append(rule)
                rule = []
                if line == "end":
                    rules[-1].append(line)
                    continue
            rule.append(line)
        result = []
        for rule in rules:
            offset = 0
            self.is_before_command_token = True
            has_continuation_line = rule[-1].startswith(" ")
            rule = "\n".join(rule) + "\n"
            while True:
                match, token_kind, offset = self.next_token(
                    rule, offset, has_continuation_line
                )
                if match is None:
                    break
                text = match.group()
                result.append((match.start(), token_kind, text))
            assert offset == len(rule), "cannot tokenize leftover text: '{}'".format(
                rule[offset:]
            )
        return result



#!/usr/bin/env python

"""Command line test driver."""

from __future__ import unicode_literals
from __future__ import print_function
import argparse
import asyncio
import datetime
from difflib import SequenceMatcher
import io
import re
import shlex
import sys
import unicodedata

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest  # type: ignore


# Directives can occur at the beginning of a line, or anywhere in a line that does not start with #.
COMMENT_RE = r"^(?:[^#].*)?#\s*"

# A regex showing how to run the file.
RUN_RE = re.compile(COMMENT_RE + r"RUN:\s+(.*)\n")
REQUIRES_RE = re.compile(COMMENT_RE + r"REQUIRES:\s+(.*)\n")

# A regex capturing lines that should be checked against stdout.
CHECK_STDOUT_RE = re.compile(COMMENT_RE + r"CHECK:\s+(.*)\n")

# A regex capturing lines that should be checked against stderr.
CHECK_STDERR_RE = re.compile(COMMENT_RE + r"CHECKERR:\s+(.*)\n")

VARIABLE_OVERRIDE_RE = re.compile(r"\w+=.*")

SKIP = object()


def find_command(program):
    import os

    path, _ = os.path.split(program)
    if path:
        return os.path.isfile(program) and os.access(program, os.X_OK)
    for path in os.environ["PATH"].split(os.pathsep):
        exe = os.path.join(path, program)
        if os.path.isfile(exe) and os.access(exe, os.X_OK):
            return exe

    return None


class Config(object):
    def __init__(self):
        # Whether to have verbose output.
        self.verbose = False
        # Whether output gets ANSI colorization.
        self.colorize = False
        # Whether to show which file was tested.
        self.progress = False

    def colors(self):
        """Return a dictionary mapping color names to ANSI escapes"""

        def ansic(n):
            return "\033[%dm" % n if self.colorize else ""

        return {
            "RESET": ansic(0),
            "BOLD": ansic(1),
            "NORMAL": ansic(39),
            "BLACK": ansic(30),
            "RED": ansic(31),
            "GREEN": ansic(32),
            "YELLOW": ansic(33),
            "BLUE": ansic(34),
            "MAGENTA": ansic(35),
            "CYAN": ansic(36),
            "LIGHTGRAY": ansic(37),
            "DARKGRAY": ansic(90),
            "LIGHTRED": ansic(91),
            "LIGHTGREEN": ansic(92),
            "LIGHTYELLOW": ansic(93),
            "LIGHTBLUE": ansic(94),
            "LIGHTMAGENTA": ansic(95),
            "LIGHTCYAN": ansic(96),
            "WHITE": ansic(97),
        }


def output(*args):
    print("".join(args) + "\n")


def esc(m):
    map = {
        "\n": "\\n",
        "\\": "\\\\",
        "\a": "\\a",
        "\b": "\\b",
        "\f": "\\f",
        "\r": "\\r",
        "\t": "\\t",
        "\v": "\\v",
    }
    if m in map:
        return map[m]
    if unicodedata.category(m)[0] == "C":
        return "\\x{:02x}".format(ord(m))
    else:
        return m


def escape_string(s):
    return "".join(esc(ch) for ch in s)


class CheckerError(Exception):
    """Exception subclass for check line parsing.

    Attributes:
      line: the Line object on which the exception occurred.
    """

    def __init__(self, message, line=None):
        super(CheckerError, self).__init__(message)
        self.line = line


class Line(object):
    """A line that remembers where it came from."""

    def __init__(self, text, number, file):
        self.text = text
        self.number = number
        self.file = file

    def __hash__(self):
        # Chosen by fair diceroll
        # No, just kidding.
        # HACK: We pass this to the Sequencematcher, which puts the Checks into a dict.
        # To force it to match the regexes, we return a hash collision intentionally,
        # so it falls back on __eq__().
        #
        # CheckCmd has the same thing.
        return 0

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, CheckCmd):
            return other.regex.match(self.text)
        if isinstance(other, Line):
            # We only compare the text here so SequenceMatcher can reshuffle these
            return self.text == other.text
        raise NotImplementedError

    def subline(self, text):
        """Return a substring of our line with the given text, preserving number and file."""
        return Line(text, self.number, self.file)

    @staticmethod
    def readfile(file, name):
        return [Line(text, idx + 1, name) for idx, text in enumerate(file)]

    def is_empty_space(self):
        return not self.text or self.text.isspace()

    def escaped_text(self, for_formatting=False):
        ret = escape_string(self.text.rstrip("\n"))
        if for_formatting:
            ret = ret.replace("{", "{{").replace("}", "}}")
        return ret


class RunCmd(object):
    """A command to run on a given Checker.

    Attributes:
        args: Unexpanded shell command as a string.
    """

    def __init__(self, args, line):
        self.args = args
        self.line = line

    @staticmethod
    def parse(line):
        if not shlex.split(line.text):
            raise CheckerError("Invalid RUN command", line)
        return RunCmd(line.text, line)


class TestFailure(object):
    def __init__(self, line, check, testrun, diff=None, lines=[], checks=[]):
        self.line = line
        self.check = check
        self.testrun = testrun
        self.error_annotation_lines = None
        self.diff = diff
        self.lines = lines
        self.checks = checks
        self.signal = None

    def message(self):
        fields = self.testrun.config.colors()
        fields["name"] = self.testrun.name
        fields["subbed_command"] = self.testrun.subbed_command
        if self.line:
            fields.update(
                {
                    "output_file": self.line.file,
                    "output_lineno": self.line.number,
                    "output_line": self.line.escaped_text(),
                }
            )
        if self.check:
            fields.update(
                {
                    "input_file": self.check.line.file,
                    "input_lineno": self.check.line.number,
                    "input_line": self.check.line.escaped_text(),
                    "check_type": self.check.type,
                }
            )
        filemsg = "" if self.testrun.config.progress else " in {name}"
        fmtstrs = ["{RED}Failure{RESET}" + filemsg + ":", ""]
        if self.signal:
            fmtstrs += [
                "  Process was killed by signal {BOLD}" + self.signal + "{RESET}",
                "",
            ]
        if self.line and self.check:
            fmtstrs += [
                "  The {check_type} on line {input_lineno} wants:",
                "    {BOLD}{input_line}{RESET}",
                "",
                "  which failed to match line {output_file}:{output_lineno}:",
                "    {BOLD}{output_line}{RESET}",
                "",
            ]

        elif self.check:
            fmtstrs += [
                "  The {check_type} on line {input_lineno} wants:",
                "    {BOLD}{input_line}{RESET}",
                "",
                "  but there was no remaining output to match.",
                "",
            ]
        else:
            fmtstrs += [
                "  There were no remaining checks left to match {output_file}:{output_lineno}:",
                "    {BOLD}{output_line}{RESET}",
                "",
            ]
        if self.error_annotation_lines:
            fields["error_annotation"] = "    ".join(
                [x.text for x in self.error_annotation_lines]
            )
            fields["error_annotation_lineno"] = str(
                self.error_annotation_lines[0].number
            )
            if len(self.error_annotation_lines) > 1:
                fields["error_annotation_lineno"] += ":" + str(
                    self.error_annotation_lines[-1].number
                )
            fmtstrs += [
                "  additional output on stderr:{error_annotation_lineno}:",
                "    {BOLD}{error_annotation}{RESET}",
            ]
        if self.diff:
            fmtstrs += ["  Context:"]
            lasthi = 0
            lastcheckline = None
            for d in self.diff.get_grouped_opcodes():
                for op, alo, ahi, blo, bhi in d:
                    color = "{BOLD}"
                    if op == "replace" or op == "delete":
                        color = "{RED}"
                    # We got a new chunk, so we print a marker.
                    if alo > lasthi:
                        fmtstrs += [
                            "    [...] from line "
                            + str(self.checks[blo].line.number)
                            + " ("
                            + self.lines[alo].file
                            + ":"
                            + str(self.lines[alo].number)
                            + "):"
                        ]
                    lasthi = ahi

                    # We print one "no more checks" after the last check and then skip any markers
                    lastcheck = False
                    for a, b in zip_longest(self.lines[alo:ahi], self.checks[blo:bhi]):
                        # Clean up strings for use in a format string - double up the curlies.
                        astr = (
                            color + a.escaped_text(for_formatting=True) + "{RESET}"
                            if a
                            else ""
                        )
                        # Convince type checker that bstr will in fact be a string when read.
                        bstr = ""
                        if b:
                            bstr = (
                                "on line "
                                + str(b.line.number)
                                + ": {BLUE}"
                                + b.line.escaped_text(for_formatting=True)
                                + "{RESET}"
                            )
                            lastcheckline = b.line.number

                        if op == "equal":
                            fmtstrs += ["    " + astr]
                        elif b and a:
                            fmtstrs += [
                                "    "
                                + astr
                                + " <= does not match "
                                + b.type
                                + " "
                                + bstr
                            ]
                        elif b:
                            fmtstrs += [
                                "    "
                                + astr
                                + " <= nothing to match "
                                + b.type
                                + " "
                                + bstr
                            ]
                        elif not b:
                            string = "    " + astr
                            if bhi == len(self.checks):
                                if not lastcheck:
                                    string += " <= no more checks"
                                    lastcheck = True
                            elif lastcheckline is not None:
                                string += (
                                    " <= no check matches this, previous check on line "
                                    + str(lastcheckline)
                                )
                            else:
                                string += " <= no check matches"
                            fmtstrs.append(string)
            fmtstrs.append("")
        fmtstrs += ["  when running command:", "    {subbed_command}"]
        return "\n".join(fmtstrs).format(**fields)

    def print_message(self):
        """Print our message to stdout."""
        print(self.message())


def perform_substitution(input_str, subs):
    """Perform the substitutions described by subs to str
    Return the substituted string.
    """
    # Sort our substitutions into a list of tuples (key, value), descending by length.
    # It needs to be descending because we need to try longer substitutions first.
    subs_ordered = sorted(subs.items(), key=lambda s: len(s[0]), reverse=True)

    def subber(m):
        # We get the entire sequence of characters.
        # Replace just the prefix and return it.
        text = m.group(1)
        for key, replacement in subs_ordered:
            if text.startswith(key):
                # shell-quote the replacement, so it's usable in #RUN lines.
                # We could loosen this and only do it for #RUN/#REQUIRES,
                # but so far we don't need it anywhere.
                return shlex.quote(replacement + text[len(key) :])
        # No substitution found, so we default to running it as-is,
        # which will end up running it via $PATH.
        return text

    return re.sub(r"%(%|[a-zA-Z0-9_-]+)", subber, input_str)


def runproc(cmd, env=None):
    """Wrapper around subprocess.Popen to save typing"""
    return asyncio.run(runproc_async(cmd, env=env))


async def runproc_async(cmd, env=None, cwd=None):
    """Wrapper around subprocess.Popen to save typing"""
    PIPE = asyncio.subprocess.PIPE
    DEVNULL = asyncio.subprocess.DEVNULL
    return await asyncio.create_subprocess_shell(
        cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, env=env, cwd=cwd
    )


class TestRun(object):
    def __init__(self, name, runcmd, checker, subs, config, env=None, cwd=None):
        self.name = name
        self.runcmd = runcmd
        self.subbed_command = perform_substitution(runcmd.args, subs)
        self.checker = checker
        self.subs = subs
        self.config = config
        self.env = env
        self.cwd = cwd

    def check(self, lines, checks):
        # Reverse our lines and checks so we can pop off the end.
        lineq = lines[::-1]
        checkq = checks[::-1]
        usedlines = []
        usedchecks = []
        mismatches = []
        while lineq and checkq:
            line = lineq[-1]
            check = checkq[-1]
            if check == line:
                # This line matched this checker, continue on.
                usedlines.append(line)
                usedchecks.append(check)
                lineq.pop()
                checkq.pop()
            elif line.is_empty_space():
                # Skip all whitespace input lines.
                lineq.pop()
            else:
                usedlines.append(line)
                usedchecks.append(check)
                mismatches.append((line, check))
                # Failed to match.
                lineq.pop()
                checkq.pop()

        # Drain empties
        while lineq and lineq[-1].is_empty_space():
            lineq.pop()

        # Store the remaining lines for the diff
        for i in lineq[::-1]:
            if not i.is_empty_space():
                usedlines.append(i)
        # Store remaining checks for the diff
        for i in checkq[::-1]:
            usedchecks.append(i)

        # If we have no more output, there's no reason to give
        # SCREENFULS of text.
        # So we truncate the check list.
        if len(usedchecks) > len(usedlines):
            usedchecks = usedchecks[: len(usedlines) + 5]

        # Do a SequenceMatch! This gives us a diff-like thing.
        diff = SequenceMatcher(a=usedlines, b=usedchecks, autojunk=False)
        # If there's a mismatch or still lines or checkers, we have a failure.
        # Otherwise it's success.
        if mismatches:
            return TestFailure(
                mismatches[0][0],
                mismatches[0][1],
                self,
                diff=diff,
                lines=usedlines,
                checks=usedchecks,
            )
        elif lineq:
            return TestFailure(
                lineq[-1], None, self, diff=diff, lines=usedlines, checks=usedchecks
            )
        elif checkq:
            return TestFailure(
                None, checkq[-1], self, diff=diff, lines=usedlines, checks=usedchecks
            )
        else:
            # Success!
            return None

    def run(self):
        """Run the command. Return a TestFailure, or None."""
        return asyncio.run(self.run_async())

    async def run_async(self):
        """Run the command. Return a TestFailure, or None."""

        def split_by_newlines(s):
            """Decode a string and split it by newlines only,
            retaining the newlines.
            """
            return [
                s + "\n"
                for s in s.decode("utf-8", errors="backslashreplace").split("\n")
            ]

        if self.config.verbose:
            print(self.subbed_command)
        proc = await runproc_async(self.subbed_command, env=self.env, cwd=self.cwd)
        stdout, stderr = await proc.communicate()

        # Work around type system limitations / bad API design which makes the typechecker unhappy.
        if proc.returncode is None:
            raise RuntimeError(
                "After `proc.communicate()` the return code must be an int."
            )
        status = proc.returncode
        # HACK: This is quite cheesy: POSIX specifies that sh should return 127 for a missing command.
        # It's also possible that it'll be returned in other situations,
        # most likely when the last command in a shell script doesn't exist.
        # So we check if the command *we execute* exists, and complain then.
        cmd = next(
            (
                word
                for word in shlex.split(self.subbed_command)
                if not VARIABLE_OVERRIDE_RE.match(word)
            )
        )
        if status == 127 and not find_command(cmd):
            raise CheckerError("Command could not be found: " + cmd)
        if status == 126 and not find_command(cmd):
            raise CheckerError("Command is not executable: " + cmd)

        outlines = [
            Line(text, idx + 1, "stdout")
            for idx, text in enumerate(split_by_newlines(stdout))
        ]
        errlines = [
            Line(text, idx + 1, "stderr")
            for idx, text in enumerate(split_by_newlines(stderr))
        ]
        outfail = self.check(outlines, self.checker.outchecks)
        errfail = self.check(errlines, self.checker.errchecks)
        # It's possible that something going wrong on stdout resulted in new
        # text being printed on stderr. If we have an outfailure, and either
        # non-matching or unmatched stderr text, then annotate the outfail
        # with it.
        if outfail and errfail and errfail.line:
            outfail.error_annotation_lines = errlines[errfail.line.number - 1 :]
            # Trim a trailing newline
            if outfail.error_annotation_lines[-1].text == "\n":
                del outfail.error_annotation_lines[-1]
        failure = outfail if outfail else errfail

        if failure and status < 0:
            # Process was killed by a signal and failed,
            # add a message.
            import signal

            # Unfortunately strsignal only exists in python 3.8+,
            # and signal.signals is 3.5+.
            if hasattr(signal, "Signals"):
                try:
                    sig = signal.Signals(-status)
                    failure.signal = sig.name + " (" + signal.strsignal(sig.value) + ")"
                except ValueError:
                    failure.signal = str(-status)
            else:
                # No easy way to get the full list,
                # make up a dict.
                signals = {
                    signal.SIGABRT: "SIGABRT",
                    signal.SIGBUS: "SIGBUS",
                    signal.SIGFPE: "SIGFPE",
                    signal.SIGILL: "SIGILL",
                    signal.SIGSEGV: "SIGSEGV",
                    signal.SIGTERM: "SIGTERM",
                }
                failure.signal = signals.get(-status, str(-status))

        return failure


class CheckCmd(object):
    def __init__(self, line, checktype, regex):
        self.line = line
        self.type = checktype
        self.regex = regex

    def __hash__(self):
        # HACK: We pass this to the Sequencematcher, which puts the Checks into a dict.
        # To force it to match the regexes, we return a hash collision intentionally,
        # so it falls back on __eq__().
        #
        # Line has the same thing.
        return 0

    def __eq__(self, other):
        # "Magical" comparison with lines and strings.
        # Typically I wouldn't use this, but it allows us to check if a line matches any check in a dict or list via
        # the `in` operator.
        if other is None:
            return False
        if isinstance(other, CheckCmd):
            return self.regex == other.regex
        if isinstance(other, Line):
            return self.regex.match(other.text)
        if isinstance(other, str):
            return self.regex.match(other)
        raise NotImplementedError

    @staticmethod
    def parse(line: Line, checktype: str) -> "CheckCmd":
        # Everything inside {{}} is a regular expression.
        # Everything outside of it is a literal string.
        # Split around {{...}}. Then every odd index will be a regex, and
        # evens will be literals.
        # Note that if {{...}} appears first we will get an empty string in
        # the split array, so the {{...}} matches are always at odd indexes.
        bracket_re = re.compile(
            r"""
                \{\{   # Two open brackets
                (.*?)  # Nongreedy capture
                \}\}   # Two close brackets
            """,
            re.VERBOSE,
        )
        pieces = bracket_re.split(line.text)
        even = True
        re_strings = []
        for i, piece in enumerate(pieces):
            if even:
                if i != 0 and piece.startswith("}"):
                    raise CheckerError(
                        "Expression ('{{%s}}}') is ambiguous, consider using '{{%s}}{{[}]}}' if you really mean that"
                        % (pieces[i - 1], pieces[i - 1]),
                        line,
                    )
                # piece is a literal string.
                re_strings.append(re.escape(piece))
            else:
                # piece is a regex (found inside {{...}}).
                # Verify the regex can be compiled.
                try:
                    re.compile(piece)
                except re.error:
                    raise CheckerError("Invalid regular expression: '%s'" % piece, line)
                re_strings.append(piece)
            even = not even
        # Enclose each piece in a non-capturing group.
        # This ensures that lower-precedence operators don't trip up catenation.
        # For example: {{b|c}}d would result in /b|cd/ which is different.
        # Backreferences are assumed to match across the entire string.
        re_strings = ["(?:%s)" % s for s in re_strings]
        # Anchor at beginning and end (allowing arbitrary whitespace), and maybe
        # a terminating newline.
        # We need the anchors because Python's match() matches an arbitrary prefix,
        # not the entire string.
        re_strings = [r"^\s*"] + re_strings + [r"\s*\n?$"]
        full_re = re.compile("".join(re_strings))
        return CheckCmd(line, checktype, full_re)


class Checker(object):
    def __init__(self, name, lines):
        self.name = name

        # Helper to yield subline containing group1 from all matching lines.
        def group1s(regex):
            for line in lines:
                m = regex.match(line.text)
                if m:
                    yield line.subline(m.group(1))

        # Find run commands.
        self.runcmds = [RunCmd.parse(sl) for sl in group1s(RUN_RE)]
        self.shebang_cmd = None
        if not self.runcmds:
            # If no RUN command has been given, fall back to the shebang.
            if lines[0].text.startswith("#!"):
                # Remove the "#!" at the beginning, and the newline at the end.
                cmd = lines[0].text[2:-1]
                self.shebang_cmd = cmd
                self.runcmds = [RunCmd(cmd + " %s", lines[0])]
            else:
                raise CheckerError("No runlines ('# RUN') found")

        self.requirecmds = [RunCmd.parse(sl) for sl in group1s(REQUIRES_RE)]

        # Find check cmds.
        self.outchecks = [
            CheckCmd.parse(sl, "CHECK") for sl in group1s(CHECK_STDOUT_RE)
        ]
        self.errchecks = [
            CheckCmd.parse(sl, "CHECKERR") for sl in group1s(CHECK_STDERR_RE)
        ]


def check_file(input_file, name, subs, config, failure_handler, env=None):
    """Check a single file. Return a True on success, False on error."""
    return asyncio.run(
        check_file_async(input_file, name, subs, config, failure_handler, env=env)
    )


async def check_file_async(
    input_file, name, subs, config, failure_handler, env=None, cwd=None
):
    """Check a single file. Return a True on success, False on error."""
    success = True
    lines = Line.readfile(input_file, name)
    checker = Checker(name, lines)

    # Run all the REQUIRES lines first,
    # if any of them fail it's a SKIP
    for reqcmd in checker.requirecmds:
        proc = await runproc_async(
            perform_substitution(reqcmd.args, subs), env=env, cwd=cwd
        )
        await proc.communicate()
        # Work around type system limitations / bad API design which makes the typechecker unhappy.
        if proc.returncode is None:
            raise RuntimeError(
                "After `proc.communicate()` the return code must be an int."
            )
        if proc.returncode > 0:
            return SKIP

    if checker.shebang_cmd is not None and not find_command(checker.shebang_cmd):
        raise CheckerError("Command could not be found: " + checker.shebang_cmd)

    # Only then run the RUN lines.
    for runcmd in checker.runcmds:
        failure = await TestRun(
            name, runcmd, checker, subs, config, env=env, cwd=cwd
        ).run_async()
        if failure:
            failure_handler(failure)
            success = False
    return success


def check_path(path, subs, config, failure_handler, env=None):
    return asyncio.run(check_path_async(path, subs, config, failure_handler, env=env))


async def check_path_async(path, subs, config, failure_handler, env=None, cwd=None):
    with io.open(path, encoding="utf-8") as fd:
        return await check_file_async(
            fd, path, subs, config, failure_handler, env=env, cwd=cwd
        )


def parse_subs(subs):
    """Given a list of input substitutions like 'foo=bar',
    return a dictionary like {foo:bar}, or exit if invalid.
    """
    result = {}
    for sub in subs:
        try:
            key, val = sub.split("=", 1)
            if not key:
                print("Invalid substitution %s: empty key" % sub)
                sys.exit(1)
            if not val:
                print("Invalid substitution %s: empty value" % sub)
                sys.exit(1)
            result[key] = val
        except ValueError:
            print("Invalid substitution %s: equal sign not found" % sub)
            sys.exit(1)
    return result


def get_argparse():
    """Return a littlecheck argument parser."""
    parser = argparse.ArgumentParser(
        description="littlecheck: command line tool tester."
    )
    parser.add_argument(
        "-s",
        "--substitute",
        type=str,
        help="Add a new substitution for RUN lines. Example: bash=/bin/bash",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-p",
        "--progress",
        action="store_true",
        dest="progress",
        help="Show the files to be checked",
        default=False,
    )
    parser.add_argument(
        "--force-color",
        action="store_true",
        dest="force_color",
        help="Force usage of color even if not connected to a terminal",
        default=False,
    )
    parser.add_argument("file", nargs="+", help="File to check")
    return parser


def main():
    args = get_argparse().parse_args()
    # Default substitution is %% -> %
    def_subs = {"%": "%"}
    def_subs.update(parse_subs(args.substitute))

    tests_count = 0
    failed = False
    skip_count = 0
    config = Config()
    config.colorize = args.force_color or sys.stdout.isatty()
    config.progress = args.progress
    fields = config.colors()

    for path in args.file:
        tests_count += 1
        fields["path"] = path
        if config.progress:
            print("Testing file {path} ... ".format(**fields), end="")
            sys.stdout.flush()
        subs = def_subs.copy()
        subs["s"] = path
        starttime = datetime.datetime.now()
        ret = check_path(path, subs, config, TestFailure.print_message)
        if ret is SKIP:
            skip_count += 1
        if not ret:
            failed = True
        elif config.progress:
            endtime = datetime.datetime.now()
            duration_ms = round((endtime - starttime).total_seconds() * 1000)
            reason = "ok"
            color = "{GREEN}"
            if ret is SKIP:
                reason = "SKIPPED"
                color = "{BLUE}"
            print(
                (color + "{reason}{RESET} ({duration} ms)").format(
                    duration=duration_ms, reason=reason, **fields
                )
            )

    # To facilitate integration with testing frameworks, use exit code 125 to indicate that all
    # tests have been skipped (primarily for use when tests are run one at a time). Exit code 125 is
    # used to indicate to automated `git bisect` runs that a revision has been skipped; we use it
    # for the same reasons git does.
    if skip_count > 0 and skip_count == tests_count:
        sys.exit(125)

    sys.exit(1 if failed else 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
sendline, expect_prompt, expect_str = sp.sendline, sp.expect_prompt, sp.expect_str
expect_prompt()

# Set up key bindings

# Movement keys from the default key bindings
home, end = "\x01", "\x05"  # Ctrl-A, Ctrl-E
left, right = "\x02", "\x06"  # Ctrl-B, Ctrl-F

# Additional keys to start selecting and dump the current selection
select, dump = "\x00", "!"  # Ctrl-Space, "!"

sendline("bind ctrl-space begin-selection")
expect_prompt()
sendline("bind ! 'echo -n \"<$(commandline --current-selection)>\"'")
expect_prompt()

# Test inclusive mode

sendline("set fish_cursor_selection_mode inclusive")

# at the beginning of the line
sendline("echo" + home + select + dump)
expect_str("<e>")
sendline("echo" + home + select + right + dump)
expect_str("<ec>")
sendline("echo" + home + right + select + left + dump)
expect_str("<ec>")

# in the middle of the line
sendline("echo" + home + right + select + dump)
expect_str("<c>")
sendline("echo" + home + right + select + right + dump)
expect_str("<ch>")
sendline("echo" + home + right + right + select + left + dump)
expect_str("<ch>")

# at the end of the line
sendline("echo" + end + select + dump)
expect_str("<>")
sendline("echo" + end + select + left + dump)
expect_str("<o>")
sendline("echo" + end + left + select + right + dump)
expect_str("<o>")

# Test exclusive mode

sendline("set fish_cursor_selection_mode exclusive")

# at the beginning of the line
sendline("echo" + home + select + dump)
expect_str("<>")
sendline("echo" + home + select + right + dump)
expect_str("<e>")
sendline("echo" + home + right + select + left + dump)
expect_str("<e>")

# in the middle of the line
sendline("echo" + home + right + select + dump)
expect_str("<>")
sendline("echo" + home + right + select + right + dump)
expect_str("<c>")
sendline("echo" + home + right + right + select + left + dump)
expect_str("<c>")

# at the end of the line
sendline("echo" + end + select + dump)
expect_str("<>")
sendline("echo" + end + select + left + dump)
expect_str("<o>")
sendline("echo" + end + left + select + right + dump)
expect_str("<o>")

# Test default setting.
# We only test that the correct implementation is chosen and rely on the detailed tests from above.

# without a configured selection mode
sendline("set -u fish_cursor_selection_mode")
sendline("echo" + home + right + select + right + dump)
expect_str("<c>")

# with an unknown setting
sendline("set fish_cursor_selection_mode unknown")
sendline("echo" + home + right + select + right + dump)
expect_str("<c>")



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

sendline("set -g fish_autosuggestion_enabled 0")
expect_prompt()
sendline("bind U,n,d,o undo; bind R,e,d,o redo")
expect_prompt()

send("echo word")
expect_str("echo word")

send("Undo")
expect_str("echo")

send("Redo\r")
expect_prompt("echo word")

sendline(
    "bind x begin-undo-group 'commandline -i \"+ \"' 'commandline -i 3' end-undo-group"
)
expect_prompt()
send("math 2 x\r")
expect_prompt("5")

send("math 4x")
send("Undo1\r")
expect_prompt("41")

send("math 5 x")
send("\x02")  # c-b, moving before the 3
send("Undo")
send("Redo")
send("9\r")  # 5 + 93
expect_prompt("98")



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc, control
import os
import platform
import sys

# Skip on macOS on Github Actions because it's too resource-starved
# and fails this a lot.
#
# Presumably we still have users on macOS that would notice binding errors
if "CI" in os.environ and platform.system() == "Darwin":
    sys.exit(127)

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

sendline("bind ctrl-l repaint")
expect_prompt()
# Clear twice (regression test for #7280).
send("\f")
expect_prompt(increment=False)
send("\f")
expect_prompt(increment=False)

# Test that kill-selection after selection is cleared doesn't crash
sendline("bind ctrl-space begin-selection")
expect_prompt()
sendline("bind ctrl-w kill-selection end-selection")
expect_prompt()
send("echo 123")
# Send Ctrl-Space using CSI u encoding
send("\x1b[32;5u")
# Send Ctrl-C to clear the command line
send("\x1b[99;5u")
# Send Ctrl-W which used to crash
send("\x1b[119;5u")
sendline("bind --erase ctrl-space ctrl-w")
expect_prompt()

# Fish should start in default-mode (i.e., emacs) bindings. The default escape
# timeout is 30ms.
#
# Because common CI systems are awful, we have to increase this:

sendline("set -g fish_escape_delay_ms 120")
expect_prompt("")

# Verify the emacs transpose word (\et) behavior using various delays,
# including none, after the escape character.

# Start by testing with no delay. This should transpose the words.
send("echo abc def")
send("\033t\r")
expect_prompt("\r\n.*def abc\r\n")  # emacs transpose words, default timeout: no delay

# Now test with a delay > 0 and < the escape timeout. This should transpose
# the words.
send("echo ghi jkl")
send("\033")
sleep(0.010)
send("t\r")
# emacs transpose words, default timeout: short delay
expect_prompt("\r\n.*jkl ghi\r\n")

# Now test with a delay > the escape timeout. The transposition should not
# occur and the "t" should become part of the text that is echoed.
send("echo mno pqr")
send("\033")
sleep(0.250)
send("t\r")
# emacs transpose words, default timeout: long delay
expect_prompt("\r\n.*mno pqrt\r\n")

# Now test that exactly the expected bind modes are defined
sendline("bind --list-modes")
expect_prompt("\r\n.*default", unmatched="Unexpected bind modes")

# Test vi key bindings.
# This should leave vi mode in the insert state.
sendline("set -g fish_key_bindings fish_vi_key_bindings")
expect_prompt()

# Go through a prompt cycle to let fish catch up, it may be slow due to ASAN
sendline("echo success: default escape timeout")
expect_prompt(
    "\r\n.*success: default escape timeout", unmatched="prime vi mode, default timeout"
)

send("echo fail: default escape timeout")
expect_str("echo fail: default escape timeout")
send("\033")

# Delay needed to allow fish to transition to vi "normal" mode. The delay is
# longer than strictly necessary to let fish catch up as it may be slow due to
# ASAN.
sleep(0.250)
send("ddi")
sendline("echo success: default escape timeout")
expect_prompt(
    "\r\n.*success: default escape timeout\r\n",
    unmatched="vi replace line, default timeout: long delay",
)

# Test replacing a single character.
send("echo TEXT")
send("\033")
# Delay needed to allow fish to transition to vi "normal" mode.
# Specifically alt+h *is* bound to __fish_man_page,
# and I have seen this think that trigger with 300ms.
#
# The next step is to rip out this test because it's much more pain than it is worth
sleep(0.400)
send("hhrAi\r")
expect_prompt(
    "\r\n.*TAXT\r\n", unmatched="vi mode replace char, default timeout: long delay"
)

# Test deleting characters with 'x'.
send("echo MORE-TEXT")
send("\033")
# Delay needed to allow fish to transition to vi "normal" mode.
sleep(0.400)
send("xxxxx\r")

# vi mode delete char, default timeout: long delay
expect_prompt(
    "\r\n.*MORE\r\n", unmatched="vi mode delete char, default timeout: long delay"
)

# Test jumping forward til before a character with t
send("echo MORE-TEXT-IS-NICE")
send("\033")
# Delay needed to allow fish to transition to vi "normal" mode.
sleep(0.250)
send("0tTD\r")

# vi mode forward-jump-till character, default timeout: long delay
expect_prompt(
    "\r\n.*MORE\r\n",
    unmatched="vi mode forward-jump-till character, default timeout: long delay",
)

# DISABLED BECAUSE IT FAILS ON GITHUB ACTIONS
# Test jumping backward til before a character with T
# send("echo MORE-TEXT-IS-NICE")
# send("\033")
# # Delay needed to allow fish to transition to vi "normal" mode.
# sleep(0.250)
# send("TSD\r")
# # vi mode backward-jump-till character, default timeout: long delay
# expect_prompt(
#     "\r\n.*MORE-TEXT-IS\r\n",
#     unmatched="vi mode backward-jump-till character, default timeout: long delay",
# )

# Test jumping backward with F and repeating
send("echo MORE-TEXT-IS-NICE")
send("\033")
# Delay needed to allow fish to transition to vi "normal" mode.
sleep(0.250)
send("F-;D\r")
# vi mode backward-jump-to character and repeat, default timeout: long delay
expect_prompt(
    "\r\n.*MORE-TEXT\r\n",
    unmatched="vi mode backward-jump-to character and repeat, default timeout: long delay",
)

# Test jumping backward with F w/reverse jump
send("echo MORE-TEXT-IS-NICE")
send("\033")
# Delay needed to allow fish to transition to vi "normal" mode.
sleep(0.250)
send("F-F-,D\r")
# vi mode backward-jump-to character, and reverse, default timeout: long delay
expect_prompt(
    "\r\n.*MORE-TEXT-IS\r\n",
    unmatched="vi mode backward-jump-to character, and reverse, default timeout: long delay",
)

# Verify that changing the escape timeout has an effect.
send("set -g fish_escape_delay_ms 100\r")
expect_prompt()

send("echo fail: lengthened escape timeout")
send("\033")
sleep(0.400)
send("ddi")
sleep(0.25)
send("echo success: lengthened escape timeout\r")
expect_prompt(
    "\r\n.*success: lengthened escape timeout\r\n",
    unmatched="vi replace line, 100ms timeout: long delay",
)

# Verify that we don't switch to vi normal mode if we don't wait long enough
# after sending escape.
send("echo fail: no normal mode")
send("\033")
sleep(0.010)
send("ddi")
send("inserted\r")
expect_prompt(
    "\r\n.*fail: no normal modediinserted\r\n",
    unmatched="vi replace line, 100ms timeout: short delay",
)

# Now set it back to speed up the tests - these don't use any escape+thing bindings!
send("set -g fish_escape_delay_ms 50\r")
expect_prompt()

# Test 't' binding that contains non-zero arity function (forward-jump) followed
# by another function (and) https://github.com/fish-shell/fish-shell/issues/2357
send("\033")
sleep(0.200)
send("ddiecho TEXT")
expect_str("echo TEXT")
send("\033")
sleep(0.200)
send("hhtTrN\r")
expect_prompt("\r\n.*TENT\r\n", unmatched="Couldn't find expected output 'TENT'")

# Test sequence key delay
send("set -g fish_sequence_key_delay_ms 200\r")
expect_prompt()
send("bind -M insert jk 'commandline -i foo'\r")
expect_prompt()
send("echo jk")
send("\r")
expect_prompt("foo")
send("echo j")
sleep(0.300)
send("k\r")
expect_prompt("jk")
send("set -e fish_sequence_key_delay_ms\r")
expect_prompt()
send("echo j")
sleep(0.300)
send("k\r")
expect_prompt("foo")


# Test '~' (togglecase-char)
# HACK: Deactivated because it keeps failing on CI
# send("\033")
# sleep(0.100)
# send("cc")
# sleep(0.50)
# send("echo some TExT\033")
# sleep(0.300)
# send("hh~~bbve~\r")
# expect_prompt("\r\n.*SOME TeXT\r\n", unmatched="Couldn't find expected output 'SOME TeXT")

send("echo echo")
send("\033")
sleep(0.200)
send("bgU\r")
expect_prompt("echo ECHO")

send("echo 125")
send("\033")
sleep(0.200)
send("0$i34\r")
expect_prompt("echo 12345")

# Test operator mode with count (d3w)
send("echo one two three four five")
send("\033")
sleep(0.200)
send("0w")
send("d3w")
sendline("")
expect_prompt("echo four five")

# Test count before operator (3dw)
send("echo one two three four five")
send("\033")
sleep(0.200)
send("0w")
send("3dw")
sendline("")
expect_prompt("echo four five")

# Test count on both (2d2w -> 4 words)
send("echo one two three four five six")
send("\033")
sleep(0.200)
send("0w")
send("2d2w")
sendline("")
expect_prompt("echo five six")

# Test change operator with count (c2w)
send("echo one two three")
send("\033")
sleep(0.200)
send("0w")
send("c2wREPLACED")
sendline("")
expect_prompt("echo REPLACED three")

# Test escape cancelling count
send("echo one two three")
send("\033")
sleep(0.200)
send("0w")
send("3")
send("\033")
sleep(0.100)
send("dw")
sendline("")
expect_prompt("echo two three")

# Now test that exactly the expected bind modes are defined
sendline("bind --list-modes")
expect_prompt(
    "F\r\nT\r\ndefault\r\nf\r\ninsert\r\noperator\r\nreplace\r\nreplace_one\r\nt\r\nvisual\r\n",
    unmatched="Unexpected vi bind modes",
)

# Test word movements
# Test 'w' with underscore - should not jump over single punctuation
send("echo abc_def")
send("\033")
sleep(0.200)
send("0wwD\r")  # From start, 'w' twice should stop at underscore, delete from there
expect_prompt(
    "\r\n.*abc\r\n", unmatched="vi mode 'w' should stop at single punctuation"
)

# Test 'w' with multiple spaces - should skip spaces and land at start of next word
send("echo abc  def")
send("\033")
sleep(0.200)
send("0wwwD\r")  # Skip 'echo', 'abc', 'def', then delete last char
expect_prompt("\r\n.*abc de\r\n", unmatched="vi mode 'w' with multiple spaces")

# Test 'w' with multiple punctuations - should stop at punctuation group
send("echo abc...def")
send("\033")
sleep(0.200)
send("0wwD\r")  # Skip 'echo', then 'w' should stop at first '.', delete to end
expect_prompt("\r\n.*abc\r\n", unmatched="vi mode 'w' with multiple punctuations")

# Test 'diw' when cursor is on space - should delete only spaces
send("echo abc  def")
send("\033")
sleep(0.200)
send("0wwhdiw\r")  # Move to 'def', back to space, delete inner word (spaces only)
expect_prompt(
    "\r\n.*abcdef\r\n", unmatched="vi mode 'diw' on space should delete spaces"
)

# Test 'daw' - should delete word and trim trailing space
send("echo abc def ghi")
send("\033")
sleep(0.200)
send("0wwdaw\r")  # Skip 'echo', move to 'def', delete word with space
expect_prompt("\r\n.*abc ghi\r\n", unmatched="vi mode 'daw' should trim trailing space")

# Test 'b' backward movement with punctuation - should stop at punctuation
send("echo abc_def")
send("\033")
sleep(0.200)
send("bD\r")  # From end, 'b' should stop at 'd', delete to end
expect_prompt("\r\n.*abc_\r\n", unmatched="vi mode 'b' should stop at punctuation")

# Test 'e' end-of-word movement
send("echo abc_def")
send("\033")
sleep(0.200)
send("0weD\r")  # From start, 'w' to 'abc', 'e' to end of 'abc', delete to end
expect_prompt("\r\n.*ab\r\n", unmatched="vi mode 'e' should move to word end")

# Test 'W' WORD movement - should skip punctuation within WORD
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send("0wWD\r")  # From start, 'w' to 'abc', 'W' should skip 'abc-def', delete 'ghi'
expect_prompt(
    "\r\n.*abc-def\r\n",
    unmatched="vi mode 'W' should treat punctuation as part of WORD",
)

# Test 'E' end-of-WORD movement
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send("0wED\r")  # From start, 'w' to 'abc', 'E' to end of 'abc-def', delete to end
expect_prompt("\r\n.*abc-de\r\n", unmatched="vi mode 'E' should move to WORD end")

# Test 'B' backward WORD movement
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send("BD\r")  # From end, 'B' backward to 'ghi', delete to end
expect_prompt("\r\n.*abc-def\r\n", unmatched="vi mode 'B' backward WORD movement")

# Test 'ge' backward to end of previous word
send("echo abc def")
send("\033")
sleep(0.200)
send("0wwgex\r")  # Move to 'def', 'ge' to 'c' of 'abc', delete char with 'x'
expect_prompt(
    "\r\n.*ab def\r\n", unmatched="vi mode 'ge' should move to previous word end"
)

# Test 'gE' backward to end of previous WORD
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send(
    "0WWgEx\r"
)  # Use 'W' to move by WORDs: to 'abc-def', then 'ghi', then 'gE' back to 'f' of 'abc-def', delete char
expect_prompt(
    "\r\n.*abc-de ghi\r\n", unmatched="vi mode 'gE' should move to previous WORD end"
)

# Test 'diW' (delete inner WORD) with punctuation
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send("0wldiW\r")  # Move to 'bc-def', delete inner WORD
expect_prompt("\r\n.*ghi\r\n", unmatched="vi mode 'diW' should delete entire WORD")

# Test 'daW' (delete a WORD) with punctuation
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send("0wldaW\r")  # Move to 'bc-def', delete a WORD with space
expect_prompt("\r\n.*ghi\r\n", unmatched="vi mode 'daW' should delete WORD and space")

# Test Unicode character category separation
# In vim, different unicode categories are separated into words
send("echo abcあいう")
send("\033")
sleep(0.200)
send("0wwD\r")  # Skip 'echo', then from 'a' of 'abc', 'w' should stop at 'あ'
expect_prompt(
    "\r\n.*abc\r\n", unmatched="vi mode 'w' should stop at Unicode category boundary"
)

# Test Unicode with multiple categories
send("echo abcあいう甲乙")
send("\033")
sleep(0.200)
send("0wwwD\r")  # Skip 'echo', 'abc', hiragana, then at kanji, delete to end
expect_prompt(
    "\r\n.*abcあいう\r\n", unmatched="vi mode 'w' should separate hiragana and kanji"
)

# Test 'cw' - change word, deletes to start of next word (like vim's 'dw')
send("echo abc def")
send("\033")
sleep(0.200)
send("0wcwXXX\r")  # Move to 'abc', 'cw' deletes 'abc', type 'XXX'
expect_prompt(
    "\r\n.*XXX def\r\n", unmatched="vi mode 'cw' should delete to start of next word"
)

# Test 'ce' - change to end of word (like vim's 'de')
send("echo abc def")
send("\033")
sleep(0.200)
send("0wceXXX\r")  # Move to 'abc', 'ce' deletes 'abc' (not the space), type 'XXX'
expect_prompt(
    "\r\n.*XXX def\r\n", unmatched="vi mode 'ce' should change to end of word"
)

# Test 'cW' - change WORD, deletes to start of next WORD (like vim's 'dW')
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send(
    "0wcWXXX\r"
)  # Move to 'abc-def', 'cW' deletes 'abc-def ' (including space), type 'XXX'
expect_prompt(
    "\r\n.*XXXghi\r\n", unmatched="vi mode 'cW' should delete to start of next WORD"
)

# Test 'cE' - change to end of WORD (like vim's 'dE')
send("echo abc-def ghi")
send("\033")
sleep(0.200)
send(
    "0wcEXXX\r"
)  # Move to 'abc-def', 'cE' deletes 'abc-def' (not the space), type 'XXX'
expect_prompt(
    "\r\n.*XXX ghi\r\n", unmatched="vi mode 'cE' should change to end of WORD"
)

# Test running commands on empty line (should not crash)
send("\033")
sleep(0.200)
send("dawdiwdwdedgedgE\r")  # run many commands
expect_prompt()

# Test accepting autosuggestions with w/W
sendline("echo test-suggestion   test-suggestion")
expect_prompt()
send("echo te")
sleep(0.100)
send("\033")  # Enter normal mode
sleep(0.200)
send("w")  # forward-word-vi should accept 'st' from autosuggestion
expect_str("echo test")
send("w")  # forward-word-vi should accept '-'
expect_str("echo test-")
send("w")  # forward-word-vi should accept 'suggestion  ' from autosuggestion
expect_str("echo test-suggestion   ")
send("W\r")  # forward-word-vi should accept 'test-suggestion' from autosuggestion
expect_prompt("test-suggestion   test-suggestion")

# Switch back to regular (emacs mode) key bindings.
sendline("set -g fish_key_bindings fish_default_key_bindings")
expect_prompt()

# Verify the custom escape timeout set earlier is still in effect.
sendline("echo fish_escape_delay_ms=$fish_escape_delay_ms")
expect_prompt(
    "\r\n.*fish_escape_delay_ms=50\r\n",
    unmatched="default-mode custom timeout not set correctly",
)

sendline("set -g fish_escape_delay_ms 200")
expect_prompt()

# Verify the emacs transpose word (\et) behavior using various delays,
# including none, after the escape character.

# Start by testing with no delay. This should transpose the words.
send("echo abc def")
send("\033")
send("t\r")
expect_prompt(
    "\r\n.*def abc\r\n", unmatched="emacs transpose words fail, 200ms timeout: no delay"
)

# Verify special characters, such as \cV, are not intercepted by the kernel
# tty driver. Rather, they can be bound and handled by fish.
sendline("bind ctrl-v 'echo ctrl-v seen'")
expect_prompt()
send("\026\r")
expect_prompt("ctrl-v seen", unmatched="ctrl-v not seen")

send("bind ctrl-o 'echo ctrl-o seen'\r")
expect_prompt()
send("\017\r")
expect_prompt("ctrl-o seen", unmatched="ctrl-o not seen")

# \x17 is ctrl-w.
send("echo git@github.com:fish-shell/fish-shell")
send("\x17\x17\r")
expect_prompt("git@github.com:", unmatched="ctrl-w does not stop at :")

send("echo git@github.com:fish-shell/fish-shell")
send("\x17\x17\x17\r")
expect_prompt("git@", unmatched="ctrl-w does not stop at @")

sendline("abbr --add foo 'echo foonanana'")
expect_prompt()
sendline("bind ' ' expand-abbr or self-insert")
expect_prompt()
send("foo ")
expect_str("echo foonanana")
send(" banana\r")
expect_str(" banana\r")
expect_prompt("foonanana banana")

# Ensure that nul can be bound properly (#3189).
send("bind ctrl-space 'echo nul seen'\r")
expect_prompt()
send("\0" * 3)
# We need to sleep briefly before emitting a newline, because when we execute the
# key bindings above we place the tty in external-proc mode (see #7483) and restoring
# the mode to shell-mode races with the newline emitted below (i.e. sometimes it may
# be echoed).
sleep(0.1)
send("\r")
expect_prompt("nul seen\r\n.*nul seen\r\n.*nul seen", unmatched="nul not seen")

# Test self-insert-notfirst. (#6603)
# Here the leading 'q's should be stripped, but the trailing ones not.
sendline("bind q self-insert-notfirst")
expect_prompt()
sendline("qqqecho qqq")
expect_prompt("qqq", unmatched="Leading qs not stripped")

# Test bigword with single-character words.
sendline("bind ctrl-g kill-bigword")
expect_prompt()
send("a b c d\x01")  # ctrl-a, move back to the beginning of the line
send("\x07")  # ctrl-g, kill bigword
sendline("echo")
expect_prompt("\n.*b c d")

# Test that overriding the escape binding works
# and does not inhibit other escape sequences (up-arrow in this case).
sendline("bind escape 'echo foo'")
expect_prompt()
send("\x1b")
expect_str("foo")
send("\x1b[A")
expect_str("bind escape 'echo foo'")
sendline("bind --erase escape")
expect_prompt()

send("    a b c d\x01")  # ctrl-a, move back to the beginning of the line
send("\x07")  # ctrl-g, kill bigword
sendline("echo")
expect_prompt("\n.*b c d")

# Check that ctrl-z can be bound
sendline('bind ctrl-z "echo bound ctrl-z"')
expect_prompt()
send("\x1a")
expect_str("bound ctrl-z")

send("echo foobar")
send("\x02\x02\x02")  # ctrl-b, backward-char
sendline("\x1bu")  # alt+u, upcase word
expect_prompt("fooBAR")

sendline("bind ctrl-z history-prefix-search-backward")
expect_prompt()
sendline("echo this continues")
expect_prompt()
send("\x1a")
sendline(" with this text")
expect_prompt("this continues with this text")

sendline(
    """
    bind ctrl-g "
        commandline --insert 'echo foo ar'
        commandline -f backward-word
        commandline --insert b
        commandline -f backward-char
        commandline -f backward-char
        commandline -f delete-char
    "
""".strip()
)
expect_prompt()
send("\x07")  # ctrl-g
send("\r")
expect_prompt("foobar")

# This should do nothing instead of crash
sendline("commandline -f backward-jump")
expect_prompt()
sendline("commandline -f self-insert")
expect_prompt()
sendline("commandline -f and")
expect_prompt()

sendline("bind ctrl-g 'sleep 1' history-pager")
expect_prompt()
send("\x07")  # ctrl-g
send("\x1b[27u")  # escape, to close pager

sendline("bind ctrl-g kill-inner-word")
expect_prompt()
send("echo foo-bar")
send("\x07")  # ctrl-g
sendline("baz")
expect_str("foo-barbaz")
expect_prompt()

sendline("bind ctrl-g kill-a-word")
expect_prompt()
send("echo foo-bar")
send("\x07")  # ctrl-g
sendline("qux")
expect_str("foo-barqux")
expect_prompt()

sendline("bind ctrl-g backward-word-end")
sendline("eco 12" + control("g") + "h")
expect_str("12")
expect_prompt()

# Check that the builtin version of `exit` works
# (for obvious reasons this MUST BE LAST)
sendline("function myexit; echo exit; exit; end; bind ctrl-z myexit")
expect_prompt()
send("\x1a")
expect_str("exit")

for t in range(0, 50):
    if not sp.spawn.isalive():
        break
    # This is cheesy, but on CI with thread-sanitizer this can be slow enough that the process is still running, so we sleep for a bit.
    sleep(0.1)
else:
    print("Fish did not exit via binding!")
    sys.exit(1)



# /// script
# requires-python = ">=3.5"
# dependencies = [
#     "launchpadlib",
# ]
# ///

from launchpadlib.launchpad import Launchpad

if __name__ == "__main__":
    launchpad = Launchpad.login_anonymously(
        "fish shell build script", "production", "~/.cache", version="devel"
    )
    ubu = launchpad.projects("ubuntu")
    print(
        "\n".join(
            x["name"]
            for x in ubu.series.entries
            if x["supported"] == True
            and x["name"] not in ("trusty", "xenial", "bionic", "focal")
        )
    )



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

sendline("fish_vi_key_bindings")
send("echo vim 1234 adsf")
send("\033")
sleep(1)
send(2 * "dge")
sendline("")
expect_re(r"\bvi\b")
expect_prompt()

sendline("echo suggest-this")
send("ech")
send("\033")
sleep(1)
send("li123")
sendline("")
expect_re(r"\bsuggest-thi123s\b")

send("echo hello world")
send("\033")
sleep(1)
send("0wdfw")
sendline("")
expect_re(r"\borld\b")

send("echo abcdef")
send("\033")
sleep(1)
send("hhhdhldl")
sendline("")
expect_re(r"\bacef\b")

# Test x copying to clipboard
send("echo hello")
send("\033")
sleep(1)
# Delete 'o', then 'l', then 'l', buffer contains 'l'
send("xxx")
# Re-insert 'l'
send("p")
sendline("")
# Results in 'hel' because 'p' pastes after the cursor ('e').
# string is 'he' + paste 'l' after 'e' = 'hel'
expect_re(r"\bhel\b")

send("echo 123456")
send("\033")
sleep(1)
send("hhh")  # cursor on 3
send("3x")  # deletes 3, 4, 5
send("p")  # pastes after 2 -> 123456
sendline("")
expect_re(r"\b123456\b")



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
import os

env = os.environ.copy()
env["TERM"] = "xterm"

sp = SpawnedProc(env=env)
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

# See that --print-colors prints the colors colored.
# Note that we don't check *all* of them, just a few.
sendline("set_color --print-colors")
expect_str("black")
expect_str("blue")
expect_str("brblack")
expect_str("brblue")
expect_str("brcyan")
expect_str("brgreen")
expect_str("brmagenta")
expect_str("brred")
expect_str("brwhite")
expect_str("bryellow")
expect_str("cyan")
expect_str("green")
expect_str("magenta")
# These should be anchored at the beginning of the line, no e.g. bold sequence before.
expect_str("\n\x1b[31mred")
expect_str("\n\x1b[37mwhite")
expect_str("\n\x1b[33myellow")
expect_str("normal")
expect_prompt()

sendline("set_color --background blue --print-colors")
expect_str("black")
expect_str("blue")
expect_str("brblack")
expect_str("brblue")
expect_str("brcyan")
expect_str("brgreen")
expect_str("brmagenta")
expect_str("brred")
expect_str("brwhite")
expect_str("bryellow")
expect_str("cyan")
expect_str("green")
expect_str("magenta")
expect_str("\x1b[31;44mred")
expect_str("white")
expect_str("yellow")
expect_str("normal")
expect_prompt()

# This used to crash
sendline("set_color --print-colors --background normal")
expect_str("black")
expect_str("blue")
expect_str("brblack")
expect_str("brblue")
expect_str("brcyan")
expect_str("brgreen")
expect_str("brmagenta")
expect_str("brred")
expect_str("brwhite")
expect_str("bryellow")
expect_str("cyan")
expect_str("green")
expect_str("magenta")
expect_str("\n\x1b[31mred")
expect_str("\n\x1b[37mwhite")
expect_str("\n\x1b[33myellow")
expect_str("normal")
expect_prompt()

# ENABLE RGB MODE - turn this off if you want to test anything after this!
sendline("set -g fish_term24bit 1")
expect_prompt()

# See that --print-colors prints the given colors.
sendline("set_color --print-colors ff0 red")
expect_str("\x1b[38;2;255;255;0mff0")
expect_str("\x1b[31mred")
expect_prompt()



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
from time import sleep
import os
import sys

# Disable under CI - keeps failing because the timing is too tight
if "CI" in os.environ:
    sys.exit(127)

os.environ["fish_escape_delay_ms"] = "10"
sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

sendline("$fish_key_reader --version")
expect_re("fish_key_reader, version .*")
expect_prompt()

sendline("$fish_key_reader -c -V")

# Do we get the expected startup prompt?
expect_str("Press a key:")

# Is a single control char echoed correctly?
send("\x07")
expect_str("# decoded from: \\x07\r\n")
expect_str("bind ctrl-g 'do something'\r\n")

# Is a non-ASCII UTF-8 sequence prefaced by an escape char handled correctly?
sleep(0.020)
send("\x1b")
expect_str("# decoded from: \\e\r\n")
expect_str("bind escape 'do something'\r\n")
send("ö")
expect_str("# decoded from: ö\r\n")
expect_str("bind ö 'do something'\r\n")
send("\u1234")
expect_str("bind ሴ 'do something'\r\n")

# Is a NULL char echoed correctly?
sleep(0.020)
send("\x00")
expect_str("bind ctrl-space 'do something'\r\n")

send("\x1b\x7f")
expect_str("# decoded from: \\e\\x7f\r\n")
expect_str("bind alt-backspace 'do something'\r\n")

send("\x1c")
expect_str(r"bind ctrl-\\ 'do something'")

# Does it keep running if handed control sequences in the wrong order?
send("\x03")
sleep(0.010)
send("\x04")

# Now send a second ctrl-d. Does that terminate the process like it should?
sleep(0.050)
send("\x04")
expect_str("Exiting at your request.\r\n")
expect_prompt()

sendline("fish_features=no-qmark-noglob $fish_key_reader")
send("?")
expect_str(r"bind \? 'do something'")
send("\x04")



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
import sys
import signal
import os

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
)


# Helper to print an error and exit.
def error_and_exit(text):
    keys = sp.colors()
    print("{RED}Test failed: {NORMAL}{TEXT}".format(TEXT=text, **keys))
    sys.exit(1)


fish_pid = sp.spawn.pid

# Launch fish_test_helper.
expect_prompt()
exe_path = os.environ.get("fish_test_helper")
if not exe_path:
    sys.exit(127)
if not os.path.exists(exe_path):
    print(f"{exe_path} does not exist")
    sys.exit(1)

sp.sendline(exe_path + " nohup_wait")

# We expect it to transfer tty ownership to fish_test_helper.
sleep(1)
tty_owner = os.tcgetpgrp(sp.spawn.child_fd)
if fish_pid == tty_owner:
    os.kill(fish_pid, signal.SIGKILL)
    error_and_exit(
        "Test failed: outer fish %d did not transfer tty owner to fish_test_helper"
        % (fish_pid)
    )


# Now we are going to tell fish to exit.
# It must not hang. But it might hang when trying to restore the tty.
os.kill(fish_pid, signal.SIGTERM)

# Loop a bit until the process exits (correct) or stops (incorrect).
# When it exits it should be due to the SIGTERM that we sent it.
for i in range(50):
    pid, status = os.waitpid(fish_pid, os.WUNTRACED | os.WNOHANG)
    if pid == 0:
        # No process ready yet, loop again.
        sleep(0.1)
    elif pid != fish_pid:
        # Some other process exited (??)
        os.kill(fish_pid, signal.SIGKILL)
        error_and_exit(
            "unexpected pid returned from waitpid. Got %d, expected %d"
            % (pid, fish_pid)
        )
    elif os.WIFSTOPPED(status):
        # Our pid stopped instead of exiting.
        # Probably it stopped because of its tcsetpgrp call during exit.
        # Kill it and report a failure.
        os.kill(fish_pid, signal.SIGKILL)
        error_and_exit("pid %d stopped instead of exiting on SIGTERM" % pid)
    elif not os.WIFSIGNALED(status):
        error_and_exit("pid %d did not signal-exit" % pid)
    elif os.WTERMSIG(status) != signal.SIGTERM:
        error_and_exit(
            "pid %d signal-exited with %d instead of %d (SIGTERM)"
            % (os.WTERMSIG(status), signal.SIGTERM)
        )
    else:
        # Success!
        sys.exit(0)
else:
    # Our loop completed without the process being returned.
    os.kill(fish_pid, signal.SIGKILL)
    error_and_exit("fish with pid %d hung after SIGTERM" % fish_pid)

# Should never get here.
os.kill(fish_pid, signal.SIGKILL)
error_and_exit("unknown, should be unreachable")



#!/usr/bin/env python3
import argparse
import asyncio
import multiprocessing
import os
import resource
import shutil
import subprocess
import sys
import tempfile
from datetime import datetime
from pathlib import Path

# TODO(python>3.8): use dict
# TODO(python>3.8): use |
from typing import Dict, Optional, Union

import littlecheck

try:
    import pexpect

    PEXPECT = True
except ImportError:
    PEXPECT = False

RESET = "\033[0m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
RED = "\033[31m"

IS_CYGWIN = os.uname().sysname.startswith(("CYGWIN_", "MSYS_"))


def makeenv(script_path: Path, home: Path) -> Dict[str, str]:
    xdg_config = home / "xdg_config_home"
    func_dir = xdg_config / "fish" / "functions"
    os.makedirs(func_dir)
    os.makedirs(xdg_config / "fish" / "conf.d")
    for func in (script_path / "test_functions").glob("*.fish"):
        shutil.copy(func, func_dir / func.parts[-1])
    shutil.copy(
        script_path / "interactive.config",
        xdg_config / "fish" / "conf.d" / "interactive.fish",
    )

    xdg_data = home / "xdg_data_home"
    os.makedirs(xdg_data)
    xdg_runtime = home / "xdg_runtime_home"
    os.makedirs(xdg_runtime)
    xdg_cache = home / "xdg_cache_home"
    os.makedirs(xdg_cache)
    tmp = home / "temp"
    os.makedirs(tmp)

    # Set up the environment variables for the test.
    env = os.environ.copy()

    # unset LANG, TERM, ...
    for var in [
        "XDG_DATA_DIRS",
        "LANGUAGE",
        "MC_SID",
        "COLORTERM",
        "KONSOLE_VERSION",
        "STY",
        "TERM",  # Erase this since we still respect TERM=dumb etc.
        "TERM_PROGRAM",
    ]:
        if var in env:
            del env[var]
    langvars = [key for key in os.environ.keys() if key.startswith("LC_")]
    for key in langvars:
        del env[key]

    env.update(
        {
            "HOME": str(home),
            "TMPDIR": str(tmp),
            "FISH_FAST_FAIL": "1",
            "FISH_UNIT_TESTS_RUNNING": "1",
            "XDG_CONFIG_HOME": str(xdg_config),
            "XDG_DATA_HOME": str(xdg_data),
            "XDG_RUNTIME_DIR": str(xdg_runtime),
            "XDG_CACHE_HOME": str(xdg_cache),
            "fish_test_helper": str(home.parent / "fish_test_helper"),
            "LANG": "C",
        }
    )

    return env


def compile_test_helper(source_path: Path, binary_path: Path) -> None:
    subprocess.run(
        [
            "cc",
            source_path,
            "-o",
            binary_path,
        ],
        check=True,
    )


def is_tmux_test(path) -> bool:
    return os.path.basename(path).startswith("tmux-")


async def main():
    if len(sys.argv) < 2:
        print("Usage: test_driver.py FISH_DIRECTORY TESTS")
        return 1

    # TODO(python>3.8): no need for abspath
    script_path = Path(os.path.abspath(__file__)).parent

    argparser = argparse.ArgumentParser(
        description="test_driver: Run fish's test suite"
    )
    argparser.add_argument(
        "--max-concurrency",
        type=int,
        help="Maximum number of tests to run concurrently. The default is the number of logical CPUs.",
        default=os.environ.get(
            "FISH_TEST_MAX_CONCURRENCY", multiprocessing.cpu_count()
        ),
    )
    argparser.add_argument(
        "fish",
        nargs=1,
        help="Directory containing fish binaries to test (typically 'target/debug')",
    )
    argparser.add_argument("file", nargs="*", help="Tests to run")
    args = argparser.parse_args()

    max_concurrency = args.max_concurrency
    if max_concurrency < 1:
        print("--max-concurrency must be at least 1")
        sys.exit(1)

    fishdir = Path(args.fish[0]).absolute()
    if not fishdir.is_dir():
        fishdir = fishdir.parent

    failcount = 0
    failed = []
    passcount = 0
    skipcount = 0
    def_subs = {"%": "%"}
    lconfig = littlecheck.Config()
    lconfig.colorize = sys.stdout.isatty()
    lconfig.progress = True

    for bin in ["fish", "fish_indent", "fish_key_reader"]:
        if os.path.exists(fishdir / bin):
            def_subs[bin] = str(fishdir / bin)
        else:
            print(f"Binary does not exist: {fishdir / bin}")
            return 127

    if args.file:
        files = [(os.path.abspath(path), path) for path in args.file]
    else:
        files = [
            (os.path.abspath(path), str(path.relative_to(Path.cwd())))
            for path in sorted(script_path.glob("checks/*.fish"))
            + sorted(script_path.glob("pexpects/*.py"))
        ]
    if os.environ.get("FISH_CI_SAN"):

        def run_in_ci_san(path) -> bool:
            if path.endswith(".py"):
                return False
            if is_tmux_test(path):
                return False
            return True

        files = [path_pair for path_pair in files if run_in_ci_san(path_pair[0])]

    if not PEXPECT and any(x.endswith(".py") for (x, _) in files):
        print(f"{RED}Skipping pexpect tests because pexpect is not installed{RESET}")

    if IS_CYGWIN and any(is_tmux_test(x) for (x, _) in files):
        print(
            f"{YELLOW}Skipping tmux tests because they are unreliable on Cygwin/MSYS{RESET}"
        )

    longest_test_name_length = max([len(arg) for _, arg in files])
    max_expected_digits_duration = 5

    def print_result(arg, result, color, duration=None, suffix=None):
        duration_str = (
            ""
            if duration is None
            else f" {str(duration_ms).rjust(max_expected_digits_duration)} ms"
        )
        suffix_str = "" if suffix is None else f"\n{suffix}"
        print(
            f"{arg.ljust(longest_test_name_length)}  {color}{result}{RESET}  {duration_str}{suffix_str}",
            flush=True,
        )

    with tempfile.TemporaryDirectory(prefix="fishtest-root-") as tmp_root:
        tmp_root = Path(tmp_root)

        compile_test_helper(
            script_path / "fish_test_helper.c",
            tmp_root / "fish_test_helper",
        )

        semaphore = asyncio.Semaphore(max_concurrency)

        async def run(f, arg) -> TestResult:
            # TODO(python>3.8): use "async with"
            if semaphore is not None:
                await semaphore.acquire()
            try:
                return await run_test(
                    tmp_root, f, arg, script_path, def_subs, lconfig, fishdir
                )
            finally:
                if semaphore is not None:
                    semaphore.release()

        tasks = [create_task(run(f, arg), name=arg) for f, arg in files]
        while tasks:
            done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                try:
                    result = await task
                except Exception as e:
                    arg = task.get_name()
                    result = TestFail(
                        arg, None, f"Test '{arg}' raised an exception: {e}"
                    )
                # TODO(python>3.8): use match statement
                if isinstance(result, TestSkip):
                    arg = result.arg
                    skipcount += 1
                    print_result(arg, "SKIPPED", BLUE)
                elif isinstance(result, TestFail):
                    # fmt: off
                    arg, duration_ms, error_message = result.arg, result.duration_ms, result.error_message
                    # fmt: on
                    failcount += 1
                    failed += [arg]
                    print_result(arg, "FAILED", RED, duration_ms, error_message)
                elif isinstance(result, TestPass):
                    arg, duration_ms = result.arg, result.duration_ms
                    passcount += 1
                    print_result(arg, "PASSED", GREEN, duration_ms)

    if passcount + failcount + skipcount > 1:
        print(f"{passcount} / {passcount + failcount} passed ({skipcount} skipped)")
    if failcount:
        failstr = "\n    ".join(failed)
        print(f"{RED}Failed tests{RESET}:\n    {failstr}")
    if passcount == 0 and failcount == 0 and skipcount:
        return 125
    return 1 if failcount else 0


# TODO(python>=3.7): @dataclass
class TestSkip:
    arg: str

    def __init__(self, arg: str):
        self.arg = arg


class TestFail:
    arg: str
    duration_ms: Optional[int]
    error_message: Optional[str]

    def __init__(
        self, arg: str, duration_ms: Optional[int], error_message: Optional[str]
    ):
        self.arg = arg
        self.duration_ms = duration_ms
        self.error_message = error_message


class TestPass:
    arg: str
    duration_ms: int

    def __init__(self, arg: str, duration_ms: int):
        self.arg = arg
        self.duration_ms = duration_ms


TestResult = Union[TestSkip, TestFail, TestPass]


# TODO(python>3.8): use asyncio.create_task
def create_task(coro, name: str) -> asyncio.Task:
    task = asyncio.Task(coro)
    if sys.version_info >= (3, 8):
        task.set_name(name)
    return task


async def run_test(
    tmp_root: Path,
    test_file_path: str,
    arg,
    script_path: Path,
    def_subs,
    lconfig,
    fishdir: Path,
) -> TestResult:
    if not test_file_path.endswith(".fish") and not test_file_path.endswith(".py"):
        return TestFail(arg, None, f"Not a valid test file: {arg}")

    starttime = datetime.now()
    home = Path(tempfile.mkdtemp(prefix="fishtest-", dir=tmp_root))
    test_env = makeenv(script_path, home)
    if IS_CYGWIN and is_tmux_test(test_file_path):
        return TestSkip(arg)
    elif test_file_path.endswith(".fish"):
        subs = def_subs.copy()
        subs.update(
            {
                "s": test_file_path,
                "fish_test_helper": str(tmp_root / "fish_test_helper"),
            }
        )

        # littlecheck
        error_message = ""

        def append_error_message(x):
            nonlocal error_message
            error_message += x.message()

        ret = await littlecheck.check_path_async(
            test_file_path,
            subs,
            lconfig,
            append_error_message,
            env=test_env,
            cwd=home,
        )
        endtime = datetime.now()
        duration_ms = round((endtime - starttime).total_seconds() * 1000)
        if ret is littlecheck.SKIP:
            return TestSkip(arg)
        elif ret:
            return TestPass(arg, duration_ms)
        else:
            return TestFail(arg, duration_ms, error_message)
    elif test_file_path.endswith(".py"):
        test_env.update(
            {
                "PYTHONPATH": str(script_path),
                "fish": str(fishdir / "fish"),
                "fish_key_reader": str(fishdir / "fish_key_reader"),
                "fish_indent": str(fishdir / "fish_indent"),
                "TERM": "dumb",
                "FISH_FORCE_COLOR": "1" if sys.stdout.isatty() else "0",
            }
        )
        if not PEXPECT:
            return TestSkip(arg)
        PIPE = asyncio.subprocess.PIPE
        proc = await asyncio.subprocess.create_subprocess_exec(
            "python3",
            test_file_path,
            stdout=PIPE,
            stderr=PIPE,
            env=test_env,
            cwd=home,
        )
        stdout, stderr = await proc.communicate()
        endtime = datetime.now()
        duration_ms = round((endtime - starttime).total_seconds() * 1000)
        returncode = proc.returncode
        if returncode == 0:
            return TestPass(arg, duration_ms)
        elif returncode == 127:
            return TestSkip(arg)
        else:
            error_message = ""
            if stdout:
                error_message += stdout.decode("utf-8")
            if stderr:
                error_message += stderr.decode("utf-8")
            return TestFail(arg, duration_ms, error_message)
    else:
        return TestFail(arg, None, "Error in test driver. This should be unreachable.")


if sys.version_info < (3, 7):

    def asyncio_run(coro):
        loop = asyncio.get_event_loop()
        try:
            return loop.run_until_complete(coro)
        finally:
            if not loop.is_closed():
                loop.close()

else:
    asyncio_run = asyncio.run

if __name__ == "__main__":
    # Increase the maximum number of open files to at least 4096,
    # as we run tests concurrently.
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    if soft < 4096:
        resource.setrlimit(resource.RLIMIT_NOFILE, (min(4096, hard), hard))
    try:
        ret = asyncio_run(main())
        sys.exit(ret)
    except KeyboardInterrupt:
        sys.exit(130)



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc, control

sp = SpawnedProc()
send, sendline, sleep, expect_prompt = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
)


def use_suggestion(*, delay=0.1):
    sleep(delay)
    send("\033[C")
    sendline("")


def run(cmd: str):
    sendline(cmd)
    expect_prompt()


expect_prompt()

run("echo Hello")

# basic samecase history suggestion
send("echo He")
use_suggestion()
expect_prompt(">echo Hello\r\nHello")

# case-correcting history suggestion
send("echo he")
use_suggestion()
expect_prompt(">echo Hello\r\nHello")

# prefer samecase history suggestions, even if older
run("echo hello")
send("echo He")
use_suggestion()
expect_prompt(">echo Hello\r\nHello")

# case-correcting command suggestion
send("Tru")
use_suggestion(delay=2.0)
expect_prompt(">true \r\n")

# the motivating example: prefer icase history suggestions over icase completion suggestions
run("mkdir -p Projects/myproject Projects/wrongproject")

# (prerequisite: without any relevant history, and with more than one subdir, fish can't suggest deeper than Projects/)
send("cd pro")
use_suggestion(delay=0.5)
expect_prompt(">cd Projects/\r\n")

run("cd ..")

# (and now the actual test)
run("cd Projects/myproject")
run("cd ../..")

send("cd pro")
use_suggestion()
expect_prompt(">cd Projects/myproject\r\n")

run("cd ../..")

# BUT prefer samecase completion suggestions over icase history suggestions
run("mkdir problems")

send("cd pro")
use_suggestion(delay=0.5)
expect_prompt(">cd problems/\r\n")

send(control("c"))
run("touch configure && chmod +x configure")
run("echo clean &&\n./configure")
run("rm configure")
send("./con")
use_suggestion()
expect_prompt(">./con\r\n")

send(control("c"))
run("touch somecommand")
run("chmod +x somecommand")
run('echo "multiline-token\n./somecommand arg1 arg2"')
send("./some")
use_suggestion()
expect_prompt(">./somecommand \r\n")



import os


def setup(app):
    current_dir = os.path.abspath(os.path.dirname(__file__))
    app.add_html_theme("python_docs_theme", current_dir)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)

from time import sleep
from subprocess import call
import os
import signal
import sys

# Disable under CI - keeps failing because the timing is too tight
if "CI" in os.environ:
    sys.exit(127)

# Test job summary for interactive shells.
expect_prompt()

sendline("function fish_job_summary; string join ':' $argv; end")
expect_prompt()

# fish_job_summary is called when background job ends.
sendline("sleep 0.5 &")
expect_prompt()
expect_re("[0-9]+:0:sleep 0.5 &:ENDED", timeout=20)
sendline("")
expect_prompt()

# fish_job_summary is called when background job is signalled.
# cmd_line correctly prints only the actually backgrounded job.
sendline("false; sleep 20 &; true")
expect_prompt()
sendline("set -l my_pid $last_pid")
expect_prompt("")
sendline("jobs")
expect_re("Job.*Group.*(CPU)?.*State.*Command")
expect_re(".*running.*sleep 20 &")
expect_prompt()
sendline("echo $my_pid")
m = expect_re("\\d+\r\n")
expect_prompt()
os.kill(int(m.group()), signal.SIGTERM)
expect_re("[0-9]+:0:sleep 20 &:SIGTERM:Polite quit request", timeout=20)
sendline("")
expect_prompt()

# fish_job_summary is called when foreground job is signalled.
# cmd_line contains the entire pipeline. proc_id and proc_name are set in a pipeline.
sendline("true | sleep 6")
sleep(0.200)
# Beware: Mac pkill requires that the -P argument come before the process name,
# else the -P argument is ignored.
call(["pkill", "-KILL", "-P", str(sp.spawn.pid), "sleep"])
expect_re("[0-9]+:1:true|sleep 6:SIGKILL:Forced quit:[0-9]+:sleep", timeout=10)



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
import os, sys

if "CI" in os.environ:
    sys.exit(127)

sp = SpawnedProc(timeout=2)
send, sendline, sleep, expect_prompt = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
)

# Verify that the fish_indent builtin can be interrupted with ctrl-c.
expect_prompt()
sendline("fish_indent")
sleep(0.1)
send("\x03")  # ctrl-c
expect_prompt()



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc
import os
import pexpect
import signal
import sys

sp = SpawnedProc()
fish_pid = sp.spawn.pid
assert fish_pid is not None
sp.expect_prompt()

# Set up a fish_exit handler that prints a marker to stdout.
sp.sendline("function __test_exit --on-event fish_exit; echo SIGTERM_CLEANUP_OK; end")
sp.expect_prompt()

# Run a command to populate history.
sp.sendline("echo sigterm_history_test_%d" % fish_pid)
sp.expect_prompt()

# Send SIGTERM.
os.kill(fish_pid, signal.SIGTERM)

# Try to catch the cleanup marker before EOF.
idx = sp.spawn.expect(["SIGTERM_CLEANUP_OK", pexpect.EOF], timeout=5)
if idx == 1:
    print("FAIL: fish_exit did not fire on SIGTERM (no cleanup marker in output)")
    sys.exit(1)

# Drain remaining output.
sp.spawn.expect(pexpect.EOF, timeout=5)
sp.spawn.close()

# Verify death was by signal (re-raised SIGTERM), not normal exit.
if sp.spawn.signalstatus != signal.SIGTERM:
    print(
        "FAIL: expected SIGTERM, got signal=%s exit=%s"
        % (sp.spawn.signalstatus, sp.spawn.exitstatus)
    )
    sys.exit(1)

# Check history file for NUL bytes.
histfile = os.path.join(os.environ["XDG_DATA_HOME"], "fish", "fish_history")
with open(histfile, "rb") as f:
    raw = f.read()
assert len(raw) != 0, "history file is empty"
nul_count = raw.count(b"\x00")
if nul_count > 0:
    print("FAIL: history file contains %d NUL bytes" % nul_count)
    sys.exit(1)



#!/usr/bin/env python3
from pexpect_helper import SpawnedProc

sp = SpawnedProc()
sendline, expect_prompt = sp.sendline, sp.expect_prompt

expect_prompt()
sendline("status job-control full")
expect_prompt()

sendline(
    "$fish -c 'status job-control full ; $fish_test_helper report_foreground' &; wait"
)
expect_prompt()

sendline("echo it worked")
expect_prompt("it worked")

# Regression test for #9181
sendline("status job-control interactive")
expect_prompt()
sendline("$fish_test_helper abandon_tty")
expect_prompt()
sendline("echo cool")
expect_prompt("cool")
sendline("true ($fish_test_helper abandon_tty)")
expect_prompt()
sendline("echo even cooler")
expect_prompt("even cooler")

sendline("TERM=not-dumb $fish -c read")
sp.send_primary_device_attribute()
sendline("something")
sp.expect_re(r"\x1b\[\?u")
expect_prompt()



#!/usr/bin/env python3
# This is a very fragile test. Sorry about that. But interactively entering
# commands and verifying they are recorded correctly in the interactive
# history and that history can be manipulated is inherently difficult.
#
# This is meant to verify just a few of the most basic behaviors of the
# interactive history to hopefully keep regressions from happening. It is not
# meant to be a comprehensive test of the history subsystem. Those types of
# tests belong in the src/fish_tests.cpp module.
#

# The history function might pipe output through the user's pager. We don't
# want something like `less` to complicate matters so force the use of `cat`.
from pexpect_helper import SpawnedProc
import os

env = os.environ.copy()
env["PAGER"] = "cat"

sp = SpawnedProc(env=env)

send, sendline, sleep, expect_prompt, expect_re, expect_str = (
    sp.send,
    sp.sendline,
    sp.sleep,
    sp.expect_prompt,
    sp.expect_re,
    sp.expect_str,
)
expect_prompt()

# ==========
# Start by ensuring we're not affected by earlier tests. Clear the history.
sendline("builtin history clear")
expect_prompt()

# ==========
# The following tests verify the behavior of the history builtin.
# ==========

# ==========
# List our history which should be empty after just clearing it.
sendline("echo start1; builtin history; echo end1")
expect_prompt("start1\r\nend1\r\n")

# Our history should now contain the previous command and nothing else.
sendline("echo start2; builtin history; echo end2")
expect_prompt("start2\r\necho start1; builtin history; echo end1\r\nend2\r\n")

# ==========
# The following tests verify the behavior of the history function.
# ==========

# ==========
# Verify explicit searching for the first two commands in the previous tests
# returns the expected results.
sendline("history search --reverse 'echo start' | cat")
expect_prompt("echo start1;.*\r\necho start2;")

# ==========
# Verify searching is the implicit action.
sendline("history -p 'echo start'")
expect_prompt("echo start2.*\r\necho start1")

# ==========
# Verify searching with a request for timestamps includes the timestamps.
sendline("history search --show-time='# %F %T%n' --prefix 'echo start'")
expect_prompt(
    "# \\d\\d\\d\\d-\\d\\d-\\d\\d \\d\\d:\\d\\d:\\d\\d\r\necho start2; .*\r\n# \\d\\d\\d\\d-\\d\\d-\\d\\d \\d\\d:\\d\\d:\\d\\d\r\necho start1;"
)

# ==========
# Verify explicit searching for an exact command returns just that command.
# returns the expected results.
sendline("echo hello")
expect_prompt()
sendline("echo goodbye")
expect_prompt()
sendline("echo hello again")
expect_prompt()
sendline("echo hello AGAIN")
expect_prompt()

sendline("history search --exact 'echo goodbye' | cat")
expect_prompt("echo goodbye\r\n")

sendline("history search --exact 'echo hello' | cat")
expect_prompt("echo hello\r\n")

# This is slightly subtle in that it shouldn't actually match anything between
# the command we sent and the next prompt.
sendline("history search --exact 'echo hell' | cat")
expect_prompt("history search --exact 'echo hell' | cat\r\n")

# Verify that glob searching works.
sendline("history search --prefix 'echo start*echo end' | cat")
expect_prompt("echo start1; builtin history; echo end1\r\n")

# ==========
# Delete a single command we recently ran.
sendline("history delete -e -C 'echo hello'")
expect_prompt()
sendline("echo count hello (history search -e -C 'echo hello' | count)")
expect_prompt("count hello 0\r\n")

# ==========
# Interactively delete one of multiple matched commands. This verifies that we
# delete the first entry matched by the prefix search (the most recent command
# sent above that matches).
sendline("history delete -p 'echo hello'")
expect_re("history delete -p 'echo hello'\r\n")
expect_re(r"\[1\] echo hello AGAIN\r\n")
expect_re(r"\[2\] echo hello again\r\n\r\n")
expect_re(
    "Enter nothing to cancel the delete, or\r\nEnter one or more of the entry IDs or ranges like '5..12', separated by a space.\r\nFor example '7 10..15 35 788..812'.\r\nEnter 'all' to delete all the matching entries.\r\n"
)
expect_re("Delete which entries\? ")
sendline("1")
expect_prompt('Deleting history entry 1: "echo hello AGAIN"\r\n')

# Verify that the deleted history entry is gone and the other one that matched
# the prefix search above is still there.
sendline("echo count AGAIN (history search -e -C 'echo hello AGAIN' | count)")
expect_prompt("count AGAIN 0\r\n")

sendline("echo count again (history search -e -C 'echo hello again' | count)")
expect_prompt("count again 1\r\n")

# Verify that the $history var has the expected content.
sendline("echo history2=$history[2]")
expect_prompt("history2=echo count AGAIN .*\r\n")

# Verify that history search is case-insensitive by default
sendline("echo term")
expect_prompt("term")
sendline("echo TERM")
expect_prompt("TERM")
sendline("echo banana")
expect_prompt("banana")
send("ter\x1b[A")  # up-arrow
expect_re("echo TERM")
sendline("")
expect_prompt("TERM")

# Check that leading space makes an ephemeral item
sendline(" echo ephemeral")
expect_prompt("ephemeral")
send("\x1b[A")  # up-arrow
expect_re(" echo ephemeral")
sendline("")
expect_prompt("ephemeral")
sendline(" ")
expect_prompt()
send("\x1b[A")
expect_re("echo TERM")  # not ephemeral!

# Verify that ctrl-l preserves an active up-arrow history search
sendline("echo __ctrl_l_history_a")
expect_prompt("__ctrl_l_history_a")
sendline("echo __ctrl_l_history_b")
expect_prompt("__ctrl_l_history_b")
send("__ctrl_l_history")
send("\x1b[A")
expect_re("echo __ctrl_l_history_b")
send("\f")
send("\x1b[A")
expect_re("echo __ctrl_l_history_a")
sendline("")
expect_prompt("__ctrl_l_history_a")

# Verify that clear-session works as expected
# Note: This test depends on that history merge resets the session from history clear-sessions point of view.
sendline("builtin history clear")
expect_prompt()

# create before history
sendline("echo before1")
expect_prompt()
sendline("echo before2")
expect_prompt()
# "reset" session with history merge
sendline("history merge")
expect_prompt()
# create after history
sendline("echo after")
expect_prompt()
# clear session
sendline("history clear-session")
expect_prompt()
sendline("history search --exact 'echo after' | cat")
expect_prompt("\r\n")

# Check history filtering
# We store anything that starts with "echo ephemeral".
sendline(
    "function fish_should_add_to_history; string match -q 'echo ephemeral*' -- $argv; and return 2; return 0; end"
)
expect_prompt("")
# Check that matching the line works
# (fish_should_add_to_history is itself stored in history so we match "ephemeral!" to avoid it)
sendline("echo ephemeral! line")
expect_prompt("ephemeral! line")
sendline("echo nonephemeral! line")
expect_prompt("nonephemeral! line")
sendline("true")
expect_prompt()
sendline("echo a; history search '*ephemeral!*' | cat; echo b")
expect_prompt(r"a\r\n.*echo nonephemeral! line\r\nb\r\n")

# If fish_should_add_to_history exists, it will completely take over,
# so even lines with spaces are stored
sendline(" echo spaced")
expect_prompt("spaced")
sendline("true")
expect_prompt()
sendline("echo a; history search '*spaced*' | cat; echo b")
expect_prompt("a\r\n.* echo spaced\r\nb\r\n")
